@@ -1,6 +1,7 @@
-from abc import ABC, abstractmethod
+from abc import ABC
 from collections import OrderedDict
 from functools import partial
+import gc
 import glob
 import json
 import os
@@ -10,7 +11,8 @@ import cv2
 import numpy as np
 from rtree import index
 
-from feabas.miscs import generate_cache, crop_image_from_bbox
+from feabas import common
+import feabas.constant as const
 
 
 # bbox :int: [xmin, ymin, xmax, ymax]
@@ -34,7 +36,7 @@ def _tile_divider_blank(imght, imgwd, x0=0, y0=0):
 def _tile_divider_border(imght, imgwd, x0=0, y0=0, cache_border_margin=10):
     """Divide image borders to cache separately. Interior is not cached."""
     divider = OrderedDict()
-    if cache_border_margin == 0:
+    if (cache_border_margin == 0) or (cache_border_margin >= min(imght//2, imgwd//2)):
         divider[1] = (x0, y0, x0+imgwd, y0+imght)
     else:
         border_ht = min(cache_border_margin, imght // 2)
@@ -79,6 +81,38 @@ def _tile_divider_block(imght, imgwd, x0=0, y0=0, cache_block_size=0):
 
 ##------------------------------ image loaders -------------------------------##
 
+def get_loader_from_json(json_info, loader_type=None, **kwargs):
+    if isinstance(json_info, str):
+        if json_info.endswith('.json'):
+            with open(json_info, 'r') as f:
+                json_obj = json.load(f)
+        elif json_info.endswith('.txt'): # could use tab separated txt, not recommend
+            if loader_type == 'StaticImageLoader':
+                loader = StaticImageLoader.from_coordinate_file(json_info)
+            else:
+                loader = MosaicLoader.from_coordinate_file(json_info)
+            json_obj = loader.init_dict()
+        else:
+            json_obj = json.loads(json_info)
+    elif isinstance(json_info, dict):
+        json_obj = json_info
+    else:
+        raise TypeError
+    json_obj.update(kwargs)
+    if loader_type is None:
+        loader_type = json_obj['ImageLoaderType']
+    if loader_type == 'DynamicImageLoader':
+        return DynamicImageLoader.from_json(json_obj)
+    elif loader_type == 'StaticImageLoader':
+        return StaticImageLoader.from_json(json_obj)
+    elif loader_type == 'MosaicLoader':
+        return MosaicLoader.from_json(json_obj)
+    elif loader_type == 'StreamLoader':
+        return StreamLoader.from_init_dict(json_obj)
+    else:
+        raise ValueError
+
+
 class AbstractImageLoader(ABC):
     """
     Abstract class for image loader.
@@ -107,8 +141,10 @@ class AbstractImageLoader(ABC):
         self._use_cache = (self._cache_size is None) or (self._cache_size > 0)
         self._init_tile_divider(**kwargs)
         self._cache_type = kwargs.get('cache_type', 'mfu')
-        self._cache = generate_cache(self._cache_type, maxlen=self._cache_size)
-        self.resolution = kwargs.get('resolution', 4.0)
+        self._cache = common.generate_cache(self._cache_type, maxlen=self._cache_size)
+        self._preprocess = kwargs.get('preprocess', None)
+        self.resolution = kwargs.get('resolution', const.DEFAULT_RESOLUTION)
+        self._read_counter = 0
 
 
     def clear_cache(self, instant_gc=False):
@@ -140,11 +176,19 @@ class AbstractImageLoader(ABC):
             self._inverse = True
 
 
-    def save_to_json(self, jsonname, **kwargs):
+    def init_dict(self, **kwargs):
         out = {'ImageLoaderType': self.__class__.__name__}
         out.update(self._export_dict(**kwargs))
-        with open(jsonname, 'w') as f:
-            json.dump(out, f, indent=2)
+        return out
+
+
+    def save_to_json(self, jsonname=None, **kwargs):
+        out = self.init_dict(**kwargs)
+        if jsonname is None:
+            return json.dumps(out, indent=2)
+        else:
+            with open(jsonname, 'w') as f:
+                json.dump(out, f, indent=2)
 
 
     def _cached_block_rtree_generator(self, bbox):
@@ -172,7 +216,7 @@ class AbstractImageLoader(ABC):
                     if dtype is None or number_of_channels is None:
                       # not sufficient info to generate empty tile, read an image to get info
                         img = self._read_image(imgpath, **kwargs)
-                        imgout = crop_image_from_bbox(img, bbox_img, bbox, return_index=False,
+                        imgout = common.crop_image_from_bbox(img, bbox_img, bbox, return_index=False,
                             return_empty=True, fillval=fillval)
                     else:
                         outht = bbox[3] - bbox[1]
@@ -210,16 +254,16 @@ class AbstractImageLoader(ABC):
                     if blk is None:
                         continue
                     if not initialized:
-                        imgp = crop_image_from_bbox(blk, blkbbox, bbox_partial,
+                        imgp = common.crop_image_from_bbox(blk, blkbbox, bbox_partial,
                             return_index=False, return_empty=True, fillval=fillval)
                         initialized = True
                     else:
-                        blkt, indx =  crop_image_from_bbox(blk, blkbbox, bbox_partial,
+                        blkt, indx =  common.crop_image_from_bbox(blk, blkbbox, bbox_partial,
                             return_index=True, return_empty=False, fillval=fillval)
                         if indx is not None and blkt is not None:
                             imgp[indx] = blkt
                 if return_index:
-                    imgout = crop_image_from_bbox(imgp, bbox_partial, bbox, return_index=True,
+                    imgout = common.crop_image_from_bbox(imgp, bbox_partial, bbox, return_index=True,
                         return_empty=return_empty, fillval=fillval)
                 else:
                     imgout = imgp
@@ -231,7 +275,7 @@ class AbstractImageLoader(ABC):
         fillval = kwargs.get('fillval', self._default_fillval)
         img = self._read_image(imgpath, **kwargs)
         bbox_img = self._get_image_bbox(imgpath)
-        imgout = crop_image_from_bbox(img, bbox_img, bbox, return_index=return_index,
+        imgout = common.crop_image_from_bbox(img, bbox_img, bbox, return_index=return_index,
             return_empty=return_empty, fillval=fillval)
         if self._use_cache:
             self._cache_image(imgpath, img=img, **kwargs)
@@ -278,8 +322,16 @@ class AbstractImageLoader(ABC):
 
     @staticmethod
     def _load_settings_from_json(jsonname):
-        with open(jsonname, 'r') as f:
-            json_obj = json.load(f)
+        if isinstance(jsonname, str):
+            if jsonname.endswith('.json'):
+                with open(jsonname, 'r') as f:
+                    json_obj = json.load(f)
+            else:
+                json_obj = json.loads(jsonname)
+        elif isinstance(jsonname, dict):
+            json_obj = jsonname
+        else:
+            raise TypeError
         settings = {}
         if 'resolution' in json_obj:
             settings['resolution'] = json_obj['resolution']
@@ -310,25 +362,24 @@ class AbstractImageLoader(ABC):
         apply_CLAHE = kwargs.get('apply_CLAHE', self._apply_CLAHE)
         inverse = kwargs.get('inverse', self._inverse)
         if number_of_channels == 3:
-            img = cv2.imread(imgpath, cv2.IMREAD_COLOR)
+            img = common.imread(imgpath, flag=cv2.IMREAD_COLOR)
         else:
-            img = cv2.imread(imgpath, cv2.IMREAD_UNCHANGED)
+            img = common.imread(imgpath, flag=cv2.IMREAD_UNCHANGED)
+        self._read_counter += 1
         if img is None:
-            invalid_image_file_error = 'Image file {} not valid!'.format(imgpath)
-            raise RuntimeError(invalid_image_file_error)
+            raise RuntimeError(f'Image file {imgpath} not valid!')
         if (number_of_channels == 1) and (len(img.shape) > 2) and (img.shape[-1] > 1):
             img = img.mean(axis=-1)
         if dtype is not None:
             img = img.astype(dtype, copy=False)
+        else:
+            dtype = img.dtype
         if apply_CLAHE:
             img = self._CLAHE.apply(img)
+        if self._preprocess is not None:
+            img = self._preprocess(img)
         if inverse:
-            if np.dtype(dtype) == np.dtype('uint8'):
-                img = 255 - img
-            elif np.dtype(dtype) == np.dtype('uint16'):
-                img = 65535 - img
-            else:
-                img = img.max() - img
+            img = common.inverse_image(img, dtype)
         return img
 
 
@@ -337,19 +388,16 @@ class AbstractImageLoader(ABC):
         return self._settings_dict(**kwargs)
 
 
-    @abstractmethod
     def _get_cached_dict(self, imgpath):
         """cached_dict get function used in self._crop_from_one_image"""
         pass
 
 
-    @abstractmethod
     def _get_image_bbox(self, imgpath):
         """image_bbox get function used in self._crop_from_one_image"""
         pass
 
 
-    @abstractmethod
     def _get_image_cached_block_rtree(self, imgpath):
         """image_cached_block_rtree get function used in self._crop_from_one_image"""
         pass
@@ -373,6 +421,11 @@ class AbstractImageLoader(ABC):
         return self._dtype
 
 
+    @property
+    def default_fillval(self):
+        return self._default_fillval
+
+
 
 class DynamicImageLoader(AbstractImageLoader):
     """
@@ -463,25 +516,44 @@ class DynamicImageLoader(AbstractImageLoader):
 class StaticImageLoader(AbstractImageLoader):
     """
     Class for image loader with predetermined image list.
-    Assuming all the images are of the same dimensions dtype and bitdepth.
+    Assuming all the images are of the same dtype and bitdepth.
 
     Args:
-        filepaths(list): fullpaths of the image file list
+        filepaths(list): fullpaths of the image file list.
+        bboxes(ndarray of int): bounding boxes of each file. If set to None,
+            assume all the tiles have the same tile size (either by reading in
+            the first image or defined by the key-word argument) and all the
+            bboxes are default to (0, 0, imgwd, imght).
     Kwargs:
-        tile_size(tuple): the size of tiles.
-            If None(default), infer from the first image encountered.
+        root_dir(str): if set, the paths in filepaths are relative path to this
+            root directory.
+        tile_size(tuple): the size of tiles. If None(default), infer from the
+            first image encountered.
+        The rest of the kwargs are passed to AbstractImageLoader constructor.
     caching format:
         self._cache[imgpath] = cache_dict{blkid: tile}
-        self._cached_block_rtree = rtree
     """
-    def __init__(self, filepaths, **kwargs):
+    def __init__(self, filepaths, bboxes=None, **kwargs):
         super().__init__(**kwargs)
-        self.imgrootdir = os.path.dirname(os.path.commonprefix(filepaths))
-        self.imgrelpaths = [os.path.relpath(s, self.imgrootdir) for s in filepaths]
+        if bboxes is None:
+            bboxes = []
+        if bool(kwargs.get('root_dir', None)):
+            self.imgrootdir = kwargs['root_dir']
+            self.imgrelpaths = filepaths
+        else:
+            self.imgrootdir = os.path.dirname(os.path.commonprefix(filepaths))
+            self.imgrelpaths = [os.path.relpath(s, self.imgrootdir) for s in filepaths]
         self.check_filename_uniqueness()
-        self._tile_size = kwargs.get('tile_size', None)
+        if len(bboxes) == 0:
+            tile_size = kwargs.get('tile_size', None)
+            if tile_size is None:
+                imgpath = os.path.join(self.imgrootdir, self.imgrelpaths[0])
+                img = super()._read_image(imgpath, number_of_channels=1, apply_CLAHE=False, inverse=False)
+                tile_size = img.shape[:2]
+            bboxes = np.tile((0, 0, tile_size[-1], tile_size[0]), (len(filepaths), 1))
+        self._file_bboxes = np.round(bboxes).astype(np.int32)
         self._cached_block_rtree = {}
-        self._divider = None
+        self._divider = {}
 
 
     def check_filename_uniqueness(self):
@@ -500,12 +572,41 @@ class StaticImageLoader(AbstractImageLoader):
     def from_json(cls, jsonname, **kwargs):
         settings, json_obj = cls._load_settings_from_json(jsonname)
         filepaths = []
+        if 'root_dir' in json_obj:
+            settings['root_dir'] = json_obj['root_dir']
         if 'tile_size' in json_obj:
             settings['tile_size'] = json_obj['tile_size']
         settings.update(kwargs)
+        filepaths = []
+        bboxes = []
+        bboxes_included = True
         for f in json_obj['images']:
             filepaths.append(f['filepath'])
-        return cls(filepaths=filepaths, **settings)
+            if bboxes_included:
+                if 'bbox' in f:
+                    bboxes.append(f['bbox'])
+                else:
+                    bboxes = False
+        return cls(filepaths=filepaths, bboxes=bboxes, **settings)
+
+
+    @classmethod
+    def from_coordinate_file(cls, filename, **kwargs):
+        imgpaths, bboxes, root_dir, resolution = common.parse_coordinate_files(filename, **kwargs)
+        kwargs.setdefault('root_dir', root_dir)
+        kwargs.setdefault('resolution', resolution)
+        return cls(filepaths=imgpaths, bboxes=bboxes, **kwargs)
+
+
+    def to_coordinate_file(self, filename, **kwargs):
+        delimiter = kwargs.get('delimiter', '\t')
+        with open(filename, 'w') as f:
+            f.write(f'{{ROOT_DIR}}{delimiter}{self.imgrootdir}\n')
+            f.write(f'{{RESOLUTION}}{delimiter}{self.resolution}\n')
+            for imgpath, bbox in zip(self.imgrelpaths, self._file_bboxes):
+                bbox_str = [str(s) for s in bbox]
+                line = delimiter.join((imgpath, *bbox_str))
+                f.write(line+'\n')
 
 
     def _cache_image(self, fileid, img=None, **kwargs):
@@ -522,7 +623,7 @@ class StaticImageLoader(AbstractImageLoader):
         else:
             cache_dict = {}
         new_cache = False
-        for bid, blkbbox in self.divider.items():
+        for bid, blkbbox in self.divider(fileid).items():
             if (bid > 0) and (bid not in cache_dict):
                 if img is None:
                     img = self._read_image(fileid, **kwargs)
@@ -535,10 +636,9 @@ class StaticImageLoader(AbstractImageLoader):
 
     def _export_dict(self, output_controls=True, cache_settings=True, image_list=True):
         out = super()._settings_dict(output_controls=output_controls, cache_settings=cache_settings)
-        if self._tile_size is not None:
-            out['tile_size'] = self._tile_size
+        out['root_dir'] = self.imgrootdir
         if image_list:
-            out['images'] = [{'filepath':s} for s in self.filepaths_generator]
+            out['images'] = [{'filepath':p, 'bbox':b.tolist()} for p, b in zip(self.imgrelpaths, self._file_bboxes)]
         return out
 
 
@@ -548,22 +648,53 @@ class StaticImageLoader(AbstractImageLoader):
                 imgpath = fileid
             else:
                 imgpath = os.path.join(self.imgrootdir, self.imgrelpaths[fileid])
-            imgpath = os.path.join(self.imgrootdir, self.imgrelpaths[fileid])
-            image_not_in_cache = '{} not in the image loader cache'.format(imgpath)
-            raise KeyError(image_not_in_cache)
+            raise KeyError(f'{imgpath} not in the image loader cache')
         else:
             return self._cache[fileid]
 
 
     def _get_image_bbox(self, fileid):
-        imght, imgwd = self.tile_size
-        return (0,0, imgwd, imght)
+        if isinstance(fileid, str):
+            filepath = fileid
+            fileid = self.fileid_lookup(filepath)
+            if fileid == -1:
+                raise KeyError(f'{filepath} not in the image loader list')
+        return self._file_bboxes[fileid]
+
+
+    def file_bboxes(self, margin=0):
+        for bbox in self._file_bboxes:
+            bbox_m = [bbox[0]-margin, bbox[1]-margin, bbox[2]+margin, bbox[3]+margin]
+            yield bbox_m
+
+
+    def _get_image_hits(self, fileid, bbox):
+        # hits[blkid] = bbox
+        imgbbox = self._get_image_bbox(fileid)
+        xmin_img, ymin_img, _, _ = imgbbox
+        bbox_normalized = (bbox[0]-xmin_img, bbox[1]-ymin_img,
+            bbox[2]-xmin_img, bbox[3]-ymin_img)
+        cached_block_rtree_normalized = self._get_image_cached_block_rtree(fileid)
+        hit_list = list(cached_block_rtree_normalized.intersection(bbox_normalized, objects=True))
+        hits = {}
+        for item in hit_list:
+            bbox_hit = item.bbox
+            bbox_out = (bbox_hit[0]+xmin_img, bbox_hit[1]+ymin_img,
+                bbox_hit[2]+xmin_img, bbox_hit[3]+ymin_img)
+            hits[item.id] = bbox_out
+        return hits
 
 
     def _get_image_cached_block_rtree(self, fileid):
-        tile_ht, tile_wd = self.tile_size
-        bbox_img = (0, 0, tile_wd, tile_ht)
-        return self.cached_block_rtree[bbox_img]
+        bbox_img = self._get_image_bbox(fileid)
+        xmin, ymin, xmax, ymax = bbox_img
+        # normalize bbox to reduce number of rtrees needed
+        imgwd, imght = round(xmax - xmin), round(ymax - ymin)
+        bbox_normalized = (0, 0, imgwd, imght)
+        if bbox_normalized not in self._cached_block_rtree:
+            self._cached_block_rtree[bbox_normalized] = index.Index(
+                    self._cached_block_rtree_generator(bbox_normalized), interleaved=True)
+        return self._cached_block_rtree[bbox_normalized]
 
 
     def _read_image(self, fileid, **kwargs):
@@ -572,16 +703,6 @@ class StaticImageLoader(AbstractImageLoader):
         else:
             imgpath = os.path.join(self.imgrootdir, self.imgrelpaths[fileid])
         img = super()._read_image(imgpath, **kwargs)
-        if self._tile_size is None:
-            self._tile_size = img.shape[:2]
-        tile_ht, tile_wd = self._tile_size
-        bbox_img = (0, 0, tile_wd, tile_ht)
-        if bbox_img not in self._cached_block_rtree:
-            self._cached_block_rtree[bbox_img] = index.Index(
-                self._cached_block_rtree_generator(bbox_img), interleaved=True)
-        if self._divider is None:
-            tile_ht, tile_wd = self._tile_size
-            self._divider = self._tile_divider(tile_ht, tile_wd, x0=0, y0=0)
         if self._dtype is None:
             self._dtype = img.dtype
         if self._number_of_channels is None:
@@ -603,18 +724,15 @@ class StaticImageLoader(AbstractImageLoader):
             return -1
 
 
-    @property
-    def cached_block_rtree(self):
-        if self._cached_block_rtree is None:
-            self._read_image(0)
-        return self._cached_block_rtree
-
-
-    @property
-    def divider(self):
-        if self._divider is None:
-            self._read_image(0)
-        return self._divider
+    def divider(self, fileid):
+        bbox_img = self._get_image_bbox(fileid)
+        xmin, ymin, xmax, ymax = bbox_img
+        # normalize bbox to reduce number of rtrees needed
+        imgwd, imght = round(xmax - xmin), round(ymax - ymin)
+        bbox_normalized = (0, 0, imgwd, imght)
+        if bbox_normalized not in self._divider:
+            self._divider[bbox_normalized] = self._tile_divider(imght, imgwd, x0=0, y0=0)
+        return self._divider[bbox_normalized]
 
 
     @property
@@ -642,13 +760,6 @@ class StaticImageLoader(AbstractImageLoader):
         return self._number_of_channels
 
 
-    @property
-    def tile_size(self):
-        if self._tile_size is None:
-            self._read_image(0)
-        return self._tile_size
-
-
 
 class MosaicLoader(StaticImageLoader):
     """
@@ -657,19 +768,21 @@ class MosaicLoader(StaticImageLoader):
     """
     def __init__(self, filepaths, bboxes, **kwargs):
         assert len(filepaths) == len(bboxes)
-        super().__init__(filepaths, **kwargs)
-        self._file_bboxes = bboxes
+        super().__init__(filepaths, bboxes, **kwargs)
         self._init_rtrees()
 
 
     @classmethod
-    def from_filepaths(cls, imgpaths, pattern='_tr(\d+)-tc(\d+)', rc_offset=[1, 1], rc_order=True, **kwargs):
+    def from_filepath(cls, imgpaths, pattern='_tr({ROW_IND}\d+)-tc({COL_IND}\d+)', **kwargs):
         """
-        pattern:    regexp pattern to parse row-column pattern
-        rc_offset:  if None, normalize
-        rc_order:   True:row-colume or False:colume-row
+        pattern: See MosaicLoader._filename_parser
+        tile_size: Size of the tile. If not provide, read-in the first image.
+        tile_offset: x-y offset added to the bboxes in the unit of tile size.
+        pixel_offset: x-y offset added to the bboxes in the unit of pixel.
         """
         tile_size = kwargs.get('tile_size', None)   # if None, read in first image
+        tile_offset = kwargs.get('tile_offset', None)
+        pixel_offset = kwargs.get('pixel_offset', None)
         if isinstance(imgpaths, str):
             if '*' in imgpaths:
                 imgpaths = glob.glob(imgpaths)
@@ -677,39 +790,24 @@ class MosaicLoader(StaticImageLoader):
                 imgpaths.sort()
             else:
                 imgpaths = [imgpaths]
-        if len(imgpaths) == 1:
-            r_c = np.array([(0, 0)])
-        else:
-            r_c = np.array([MosaicLoader._filename_parser(s, pattern, rc_order=rc_order) for s in imgpaths])
-            if rc_offset is None:
-                r_c -= r_c.min(axis=0)
-            else:
-                r_c -= np.array(rc_offset).reshape(1,2)
         if tile_size is None:
-            img = cv2.imread(imgpaths[0], cv2.IMREAD_UNCHANGED)
+            img = common.imread(imgpaths[0], flag=cv2.IMREAD_UNCHANGED)
             imght, imgwd = img.shape[0], img.shape[1]
             tile_size = (imght, imgwd)
         bboxes = []
-        for rc in r_c:
-            r = rc[0]
-            c = rc[1]
-            bbox = [c*tile_size[1], r*tile_size[0], (c+1)*tile_size[1], (r+1)*tile_size[0]]
+        for fname in imgpaths:
+            bbox = MosaicLoader._filename_parser(fname, pattern, tile_size)
+            if tile_offset is not None:
+                dx = tile_offset[0] * tile_size[-1]
+                dy = tile_offset[-1] * tile_size[0] 
+                bbox = (bbox[0]+dx, bbox[1]+dy, bbox[2]+dx, bbox[3]+dy)
+            if pixel_offset is not None:
+                dx, dy = pixel_offset
+                bbox = (bbox[0]+dx, bbox[1]+dy, bbox[2]+dx, bbox[3]+dy)
             bboxes.append(bbox)
         return cls(filepaths=imgpaths, bboxes=bboxes, **kwargs)
 
 
-    @classmethod
-    def from_json(cls, jsonname, **kwargs):
-        settings, json_obj = cls._load_settings_from_json(jsonname)
-        settings.update(kwargs)
-        filepaths = []
-        bboxes = []
-        for f in json_obj['images']:
-            filepaths.append(f['filepath'])
-            bboxes.append(f['bbox'])
-        return cls(filepaths=filepaths, bboxes=bboxes, **settings)
-
-
     def _file_rtree_generator(self):
         """
         generator function to build rtree of image bounding boxes.
@@ -731,7 +829,7 @@ class MosaicLoader(StaticImageLoader):
                     blk, indx = super().crop(bbox, fileid, return_empty=return_empty, return_index=True, **kwargs)
                     if blk is not None:
                         out[indx] = blk
-        else:
+        if not initialized:
             if return_empty:
                 out = super().crop(bbox, 0, return_empty=True, return_index=False, **kwargs)
             else:
@@ -739,57 +837,6 @@ class MosaicLoader(StaticImageLoader):
         return out
 
 
-    def file_bboxes(self, margin=0):
-        for bbox in self._file_bboxes:
-            bbox_m = [bbox[0]-margin, bbox[1]-margin, bbox[2]+margin, bbox[3]+margin]
-            yield bbox_m
-
-
-    def _export_dict(self, output_controls=True, cache_settings=True, image_list=True):
-        out = super()._settings_dict(output_controls=output_controls, cache_settings=cache_settings)
-        if image_list:
-            out['images'] = [{'filepath':p, 'bbox':b} for p, b in zip(self.filepaths_generator, self._file_bboxes)]
-        return out
-
-
-    def _get_image_bbox(self, fileid):
-        if isinstance(fileid, str):
-            filepath = fileid
-            fileid = self.fileid_lookup(filepath)
-            if fileid == -1:
-                image_not_in_list = '{} not in the image loader list'.format(filepath)
-                raise KeyError(image_not_in_list)
-        return self._file_bboxes[fileid]
-
-
-    def _get_image_cached_block_rtree(self, fileid):
-        bbox_img = self._get_image_bbox(fileid)
-        xmin, ymin, xmax, ymax = bbox_img
-        # normalize bbox to reduce number of rtrees needed
-        bbox_normalized = (0, 0, xmax - xmin, ymax - ymin)
-        if bbox_normalized not in self._cached_block_rtree:
-            self._cached_block_rtree[bbox_normalized] = index.Index(
-                    self._cached_block_rtree_generator(bbox_normalized), interleaved=True)
-        return self._cached_block_rtree[bbox_normalized]
-
-
-    def _get_image_hits(self, fileid, bbox):
-        # hits[blkid] = bbox
-        imgbbox = self._get_image_bbox(fileid)
-        xmin_img, ymin_img, _, _ = imgbbox
-        bbox_normalized = (bbox[0]-xmin_img, bbox[1]-ymin_img,
-            bbox[2]-xmin_img, bbox[3]-ymin_img)
-        cached_block_rtree_normalized = self._get_image_cached_block_rtree(fileid)
-        hit_list = list(cached_block_rtree_normalized.intersection(bbox_normalized, objects=True))
-        hits = {}
-        for item in hit_list:
-            bbox_hit = item.bbox
-            bbox_out = (bbox_hit[0]+xmin_img, bbox_hit[1]+ymin_img,
-                bbox_hit[2]+xmin_img, bbox_hit[3]+ymin_img)
-            hits[item.id] = bbox_out
-        return hits
-
-
     def _init_rtrees(self):
         self._file_rtree = index.Index(self._file_rtree_generator(), interleaved=True)
         self._cache_block_rtree = {}
@@ -804,17 +851,195 @@ class MosaicLoader(StaticImageLoader):
 
 
     @ staticmethod
-    def _filename_parser(fname, pattern, rc_order=True):
+    def _filename_parser(fname, pattern, tile_size):
+        """
+        given a filename and pattern, return the bboxes. pattern str follows
+        regular expression, expect the following reserved keywords, which are
+        used to indicate the nature of the field and the will be removed during
+        parsing.
+            {ROW_IND}: row_index
+            {COL_IND}: col_index
+            {X_MIN}: minimum of x coordinates
+            {Y_MIN}: minimum of y coordinates
+            {X_MAX}: maximum of x coordinates
+            {Y_MAX}: maximum of y coordinates
+        each keyword should only appear once and has one-to-one correspondance
+        with a group in regexp.
+        e.g. feabas_tr1_is_not_a_pokemon_tc2.png can be parsed with pattern:
+            _tr({ROW_IND}\d+)\w+_tc({COL_IND}\d+)
+        """
+        keywords = ['{ROW_IND}','{COL_IND}','{X_MIN}','{Y_MIN}','{X_MAX}','{Y_MAX}']
+        pos = np.array([pattern.find(s) for s in keywords])
+        kw_pos = sorted([(p, s) for p, s in zip(pos, keywords) if p >= 0])
+        used_keywords = [s[1] for s in kw_pos]
+        for kw in used_keywords:
+            pattern = pattern.replace(kw, '')
         m = re.findall(pattern, fname)
-        if rc_order:
-            r = int(m[0][0])
-            c = int(m[0][1])
+        grps = {}
+        for k, kw in enumerate(used_keywords):
+            grps[kw] = int(m[0][k])
+        if ('{X_MIN}' in grps) and ('{X_MAX}' in grps):
+            xmin, xmax = grps['{X_MIN}'], grps['{X_MAX}']
+        elif '{X_MIN}' in grps:
+            xmin = grps['{X_MIN}']
+            xmax = xmin + tile_size[-1]
+        elif '{X_MAX}' in grps:
+            xmax = grps['{X_MAX}']
+            xmin = xmax - tile_size[-1]
+        elif '{COL_IND}' in grps:
+            xmin = grps['{COL_IND}'] * tile_size[-1]
+            xmax = xmin + tile_size[-1]
         else:
-            c = int(m[0][0])
-            r = int(m[0][1])
-        return r, c
+            raise RuntimeError(f'x position of tile not defined in filename {fname}')
+        if ('{Y_MIN}' in grps) and ('{Y_MAX}' in grps):
+            ymin, ymax = grps['{Y_MIN}'], grps['{Y_MAX}']
+        elif '{Y_MIN}' in grps:
+            ymin = grps['{Y_MIN}']
+            ymax = ymin + tile_size[0]
+        elif '{Y_MAX}' in grps:
+            ymax = grps['{Y_MAX}']
+            ymin = ymax - tile_size[0]
+        elif '{ROW_IND}' in grps:
+            ymin = grps['{ROW_IND}'] * tile_size[0]
+            ymax = ymin + tile_size[0]
+        else:
+            raise RuntimeError(f'y position of tile not defined in filename {fname}')
+        return (xmin, ymin, xmax, ymax)
 
 
     @property
     def bounds(self):
         return self._file_rtree.bounds
+
+
+
+class StreamLoader(AbstractImageLoader):
+    """
+    Loader class for images already in RAM. Should mimic the interface of that
+    of MosaicLoader.
+    """
+    def __init__(self, img, **kwargs):
+        self._img = img
+        self._dtype = kwargs.get('dtype', img.dtype)  
+        self._apply_CLAHE = kwargs.get('apply_CLAHE', False)
+        if len(img.shape) < 3:
+            self._src_number_of_channels = 0
+        else:
+            self._src_number_of_channels = img.shape[-1]
+        self._number_of_channels = kwargs.get('number_of_channels', max(1, self._src_number_of_channels))
+        self._clahe_img = None
+        self._preprocess = kwargs.get('preprocess', None)
+        self._inverse = kwargs.get('inverse', False)
+        self._default_fillval = kwargs.get('fillval', 0)
+        self.resolution = kwargs.get('resolution', const.DEFAULT_RESOLUTION)
+        self.x0 = kwargs.get('x0', 0)
+        self.y0 = kwargs.get('y0', 0)
+
+
+    @classmethod
+    def from_filepath(cls, imgpath, **kwargs):
+        img = common.imread(imgpath, flag=cv2.IMREAD_UNCHANGED)
+        return cls(img, **kwargs)
+
+
+    @classmethod
+    def from_init_dict(cls, init_dict):
+        img = init_dict.pop('img')
+        return cls(img, **init_dict)
+
+
+    def crop(self, bbox, return_empty=False, **kwargs):
+        fillval = kwargs.get('fillval', self._default_fillval)
+        bbox_img = self.bounds
+        img = self._read_image(**kwargs)
+        return common.crop_image_from_bbox(img, bbox_img, bbox,
+            return_empty=return_empty, return_index=False, fillval=fillval)
+
+
+    @property
+    def clahe_img(self):
+        if self._clahe_img is None:
+            clahe_filter = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8,8))
+            self._clahe_img = clahe_filter.apply(self.img)
+        return self._clahe_img
+
+
+    def _read_image(self, **kwargs):
+        number_of_channels = kwargs.get('number_of_channels', self._number_of_channels)
+        if number_of_channels is None:
+            number_of_channels = self._src_number_of_channels
+        dtype = kwargs.get('dtype', self._dtype)
+        apply_CLAHE = kwargs.get('apply_CLAHE', self._apply_CLAHE)
+        inverse = kwargs.get('inverse', self._inverse)
+        if apply_CLAHE:
+            img = self.clahe_img
+        else:
+            img = self._img
+        if (self._src_number_of_channels <= 1) and (number_of_channels > 1):
+            img = np.tile(img.reshape(img.shape[0], img.shape[1], 1), (1,1,number_of_channels))
+        elif (self._src_number_of_channels > 1) and (number_of_channels == 1):
+            img = img.mean(axis=-1)
+        if (dtype is not None) and (np.dtype(dtype) != img.dtype):
+            img = img.astype(dtype)
+        if self._preprocess is not None:
+            img = self._preprocess(img)
+        if inverse:
+            img = common.inverse_image(img, dtype)
+        return img
+
+
+    def clear_cache(self, instant_gc=False):
+        self._clahe_img = None
+        if instant_gc:
+            gc.collect()
+
+
+    def init_dict(self, **kwargs):
+        return super().init_dict(**kwargs)
+
+
+    def save_to_json(self, jsonname, **kwargs):
+        raise NotImplementedError
+
+
+    def _export_dict(self):
+        out = super()._export_dict(output_controls=True, cache_settings=False)
+        out['img'] = self._img
+        out['x0'] = self.x0
+        out['y0'] = self.y0
+        return out
+
+
+    def file_bboxes(self, margin=0):
+        bbox = [self.x0 - margin,
+                self.y0 - margin,
+                self.x0 + self._img.shape[1] + margin,
+                self.y0 + self._img.shape[0] + margin]
+        return [bbox]
+
+
+    @property
+    def bounds(self):
+        return self.file_bboxes(margin=0)[0]
+
+
+
+class MultiResolutionImageLoader:
+    """
+    A collection of image loaders with different resolution. Override __getitem__
+    method to automatically select the right loader for the right resolution.
+    """
+    def __init__(self, loaders, overkill=True):
+        self._loaders = loaders
+        self._resolution = np.array([ldr.resolution for ldr in loaders])
+        self._overkill = overkill # whether to prioritize higher resolution
+
+
+    def __getitem__(self, resolution):
+        dis = resolution - self._resolution
+        if self._overkill:
+            dis[dis<0] = np.abs(dis[dis<0]) + dis.max()
+        else:
+            dis = np.abs(dis)
+        indx = np.argmin(dis)
+        return self._loaders[indx]
